@@ -1,6 +1,6 @@
 '''
 
-Copyright 2020 Ryan (Mohammad) Solgi
+Copyright 2020 Ryan (Mohammad) Solgi, Demetry Pascal
 
 Permission is hereby granted, free of charge, to any person obtaining a copy of 
 this software and associated documentation files (the "Software"), to deal in 
@@ -26,10 +26,13 @@ SOFTWARE.
 ###############################################################################
 ###############################################################################
 
-import numpy as np
 import sys
 import time
+
+import numpy as np
+
 from func_timeout import func_timeout, FunctionTimedOut
+
 import matplotlib.pyplot as plt
 
 ###############################################################################
@@ -44,7 +47,6 @@ class geneticalgorithm2():
     continuous, integers, or mixed variables.
     
     
-    
     Implementation and output:
         
         methods:
@@ -64,7 +66,6 @@ class geneticalgorithm2():
                  variable_boundaries=None,\
                  variable_type_mixed=None, \
                  function_timeout=10,\
-                 no_plot = False,
                  algorithm_parameters={'max_num_iteration': None,\
                                        'population_size':100,\
                                        'mutation_probability':0.1,\
@@ -107,8 +108,6 @@ class geneticalgorithm2():
         output before function_timeout (unit is seconds) the algorithm raise error.
         For example, when there is an infinite loop in the given function. 
         
-        @param no_plot <boolean> - do not plot results using matplotlib
-
         @param algorithm_parameters:
             @ max_num_iteration <int> - stoping criteria of the genetic algorithm (GA)
             @ population_size <int> 
@@ -116,24 +115,21 @@ class geneticalgorithm2():
             @ elit_ration <float in [0,1]>
             @ crossover_probability <float in [0,1]>
             @ parents_portion <float in [0,1]>
-            @ crossover_type <string> - Default is 'uniform'; 'one_point' or 
-            'two_point' are other options
-            @ max_iteration_without_improv <int> - maximum number of 
-            successive iterations without improvement. If None it is ineffective
+            @ crossover_type <string> - Default is 'uniform'; 'one_point' or 'two_point' are other options
+            @ max_iteration_without_improv <int> - maximum number of successive iterations without improvement. If None it is ineffective
         
         for more details and examples of implementation please visit:
-            https://github.com/rmsolgi/geneticalgorithm
+            https://github.com/PasaOpasen/geneticalgorithm2
   
         '''
-        self.__name__=geneticalgorithm2
-        self.plot = not no_plot
+        self.__name__ = geneticalgorithm2
         self.report = []
 
         #############################################################
         # input function
-        assert (callable(function)),"function must be callable"     
+        assert (callable(function)), "function must be callable!"     
         
-        self.f=function
+        self.f = function
         #############################################################
         #dimension
         
@@ -142,39 +138,38 @@ class geneticalgorithm2():
         #############################################################
         # input variable type
         
-        assert(variable_type=='bool' or variable_type=='int' or\
-               variable_type=='real'), \
-               "\n variable_type must be 'bool', 'int', or 'real'"
+        assert(variable_type in ['bool', 'int', 'real']), \
+               f"\n variable_type must be 'bool', 'int', or 'real', got {variable_type}"
        #############################################################
         # input variables' type (MIXED)     
 
         if variable_type_mixed is None:
             
-            if variable_type=='real': 
-                self.var_type=np.array([['real']]*self.dim)
+            if variable_type == 'real': 
+                self.var_type = np.array([['real']]*self.dim)
             else:
-                self.var_type=np.array([['int']]*self.dim)            
+                self.var_type = np.array([['int']]*self.dim)            
 
  
         else:
-            assert (type(variable_type_mixed).__module__=='numpy'),\
+            assert (type(variable_type_mixed).__module__ == 'numpy'),\
             "\n variable_type must be numpy array"  
             assert (len(variable_type_mixed) == self.dim), \
             "\n variable_type must have a length equal dimension."       
 
             for i in variable_type_mixed:
-                assert (i=='real' or i=='int'),\
+                assert (i == 'real' or i == 'int'),\
                 "\n variable_type_mixed is either 'int' or 'real' "+\
                 "ex:['int','real','real']"+\
                 "\n for 'boolean' use 'int' and specify boundary as [0,1]"
                 
 
-            self.var_type=variable_type_mixed
+            self.var_type = variable_type_mixed
         #############################################################
         # input variables' boundaries 
 
             
-        if variable_type!='bool' or type(variable_type_mixed).__module__=='numpy':
+        if variable_type != 'bool' or type(variable_type_mixed).__module__=='numpy':
                        
             assert (type(variable_boundaries).__module__=='numpy'),\
             "\n variable_boundaries must be numpy array"
@@ -188,31 +183,31 @@ class geneticalgorithm2():
                 "\n boundary for each variable must be a tuple of length two." 
                 assert(i[0]<=i[1]),\
                 "\n lower_boundaries must be smaller than upper_boundaries [lower,upper]"
-            self.var_bound=variable_boundaries
+            self.var_bound = variable_boundaries
         else:
-            self.var_bound=np.array([[0,1]]*self.dim)
+            self.var_bound = np.array([[0,1]]*self.dim)
  
         ############################################################# 
         #Timeout
-        self.funtimeout=float(function_timeout)
+        self.funtimeout = float(function_timeout)
         
         ############################################################# 
         # input algorithm's parameters
         
-        self.param=algorithm_parameters
+        self.param = algorithm_parameters
         
-        self.pop_s=int(self.param['population_size'])
+        self.pop_s = int(self.param['population_size'])
         
         assert (self.param['parents_portion']<=1\
                 and self.param['parents_portion']>=0),\
         "parents_portion must be in range [0,1]" 
         
-        self.par_s=int(self.param['parents_portion']*self.pop_s)
-        trl=self.pop_s-self.par_s
+        self.par_s = int(self.param['parents_portion']*self.pop_s)
+        trl= self.pop_s - self.par_s
         if trl % 2 != 0:
             self.par_s+=1
                
-        self.prob_mut=self.param['mutation_probability']
+        self.prob_mut = self.param['mutation_probability']
         
         assert (self.prob_mut<=1 and self.prob_mut>=0), \
         "mutation_probability must be in range [0,1]"
@@ -225,194 +220,200 @@ class geneticalgorithm2():
         assert (self.param['elit_ratio']<=1 and self.param['elit_ratio']>=0),\
         "elit_ratio must be in range [0,1]"                
         
-        trl=self.pop_s*self.param['elit_ratio']
+        trl = self.pop_s*self.param['elit_ratio']
         if trl<1 and self.param['elit_ratio']>0:
             self.num_elit=1
         else:
-            self.num_elit=int(trl)
+            self.num_elit = int(trl)
             
         assert(self.par_s>=self.num_elit), \
         "\n number of parents must be greater than number of elits"
         
-        if self.param['max_num_iteration']==None:
-            self.iterate=0
+        if self.param['max_num_iteration'] == None:
+            self.iterate = 0
             for i in range (0,self.dim):
                 if self.var_type[i]=='int':
-                    self.iterate+=(self.var_bound[i][1]-self.var_bound[i][0])*self.dim*(100/self.pop_s)
+                    self.iterate += (self.var_bound[i][1]-self.var_bound[i][0])*self.dim*(100/self.pop_s)
                 else:
-                    self.iterate+=(self.var_bound[i][1]-self.var_bound[i][0])*50*(100/self.pop_s)
-            self.iterate=int(self.iterate)
+                    self.iterate += (self.var_bound[i][1]-self.var_bound[i][0])*50*(100/self.pop_s)
+            self.iterate = int(self.iterate)
             if (self.iterate*self.pop_s)>10000000:
                 self.iterate=10000000/self.pop_s
         else:
-            self.iterate=int(self.param['max_num_iteration'])
+            self.iterate = int(self.param['max_num_iteration'])
         
-        self.c_type=self.param['crossover_type']
-        assert (self.c_type=='uniform' or self.c_type=='one_point' or\
-                self.c_type=='two_point'),\
+        self.c_type = self.param['crossover_type']
+        assert (self.c_type in ['uniform','one_point','two_point']),\
         "\n crossover_type must 'uniform', 'one_point', or 'two_point' Enter string" 
         
         
-        self.stop_mniwi=False
-        if self.param['max_iteration_without_improv']==None:
-            self.mniwi=self.iterate+1
+        self.stop_mniwi=False ## what
+        if self.param['max_iteration_without_improv'] == None or self.param['max_iteration_without_improv'] < 1:
+            self.mniwi = self.iterate+1
         else: 
-            self.mniwi=int(self.param['max_iteration_without_improv'])
+            self.mniwi = int(self.param['max_iteration_without_improv'])
 
         
         ############################################################# 
-    def run(self):
-        
-        
+    def run(self, no_plot = False):
+        """
+        @param no_plot <boolean> - do not plot results using matplotlib by default
+        """
         ############################################################# 
         # Initial Population
         
-        self.integers=np.where(self.var_type=='int')
-        self.reals=np.where(self.var_type=='real')
+        self.integers = np.where(self.var_type == 'int')
+        self.reals = np.where(self.var_type == 'real')
         
         
         
-        pop=np.array([np.zeros(self.dim+1)]*self.pop_s)
-        solo=np.zeros(self.dim+1)
-        var=np.zeros(self.dim)       
+        pop = np.empty((self.pop_s, self.dim+1)) #np.array([np.zeros(self.dim+1)]*self.pop_s)
+        solo = np.empty(self.dim+1)
+        var = np.empty(self.dim)       
         
-        for p in range(0,self.pop_s):
+        for p in range(0, self.pop_s):
          
             for i in self.integers[0]:
-                var[i]=np.random.randint(self.var_bound[i][0],\
-                        self.var_bound[i][1]+1)  
-                solo[i]=var[i].copy()
+                var[i] = np.random.randint(self.var_bound[i][0],self.var_bound[i][1]+1)  
+                solo[i] = var[i]#.copy()
+            
             for i in self.reals[0]:
-                var[i]=self.var_bound[i][0]+np.random.random()*\
-                (self.var_bound[i][1]-self.var_bound[i][0])    
-                solo[i]=var[i].copy()
+                var[i] = np.random.uniform(self.var_bound[i][0], self.var_bound[i][1]) #self.var_bound[i][0]+np.random.random()*(self.var_bound[i][1]-self.var_bound[i][0])    
+                solo[i] = var[i]#.copy()
 
 
-            obj=self.sim(var)            
-            solo[self.dim]=obj
-            pop[p]=solo.copy()
+            obj = self.sim(var)  # simulation returns exception or func value           
+            solo[self.dim] = obj
+            pop[p] = solo.copy()
 
         #############################################################
 
         #############################################################
         # Report
-        self.report=[]
-        self.test_obj=obj
-        self.best_variable=var.copy()
-        self.best_function=obj
+        self.report = []
+        self.test_obj = obj
+        self.best_variable = var.copy()
+        self.best_function = obj
         ##############################################################   
                         
         t=1
-        counter=0
-        while t<=self.iterate:
+        counter = 0
+        while t <= self.iterate:
             
             
-            self.progress(t,self.iterate,status="GA is running...")
+            self.progress(t, self.iterate, status = f"GA is running...{t} from {self.iterate}")
             #############################################################
             #Sort
             pop = pop[pop[:,self.dim].argsort()]
 
                 
-            
-            if pop[0,self.dim]<self.best_function:
-                counter=0
-                self.best_function=pop[0,self.dim].copy()
-                self.best_variable=pop[0,: self.dim].copy()
+            if pop[0,self.dim] < self.best_function:
+                counter = 0
+                self.best_function=pop[0, self.dim]#.copy()
+                self.best_variable=pop[0,:self.dim].copy()
+                
+                self.progress(t, self.iterate, status = f"GA is running...{t} from {self.iterate}...best value = {self.best_function}")
             else:
-                counter+=1
+                counter += 1
             #############################################################
             # Report
 
-            self.report.append(pop[0,self.dim])
+            self.report.append(pop[0, self.dim])
     
             ##############################################################         
             # Normalizing objective function 
             
-            normobj=np.zeros(self.pop_s)
+            #normobj=  np.zeros(self.pop_s)
             
-            minobj=pop[0,self.dim]
-            if minobj<0:
-                normobj=pop[:,self.dim]+abs(minobj)
+            minobj = pop[0, self.dim]
+            if minobj < 0:
+                normobj=pop[:,self.dim] - minobj #+abs(minobj)
                 
             else:
                 normobj=pop[:,self.dim].copy()
     
-            maxnorm=np.amax(normobj)
-            normobj=maxnorm-normobj+1
+            maxnorm = np.amax(normobj)
+            normobj = (maxnorm + 1) - normobj 
 
             #############################################################        
             # Calculate probability
             
-            sum_normobj=np.sum(normobj)
-            prob=np.zeros(self.pop_s)
-            prob=normobj/sum_normobj
-            cumprob=np.cumsum(prob)
+            sum_normobj = np.sum(normobj)
+            prob = normobj/sum_normobj
+            cumprob = np.cumsum(prob)
   
             #############################################################        
             # Select parents
-            par=np.array([np.zeros(self.dim+1)]*self.par_s)
+            par = np.empty((self.par_s, self.dim+1))  #np.array([np.zeros(self.dim+1)]*self.par_s)
             
-            for k in range(0,self.num_elit):
+            for k in range(0, self.num_elit):
                 par[k]=pop[k].copy()
-            for k in range(self.num_elit,self.par_s):
-                index=np.searchsorted(cumprob,np.random.random())
-                par[k]=pop[index].copy()
+            # it can be vectorized
+            for k in range(self.num_elit, self.par_s):
+                index = np.searchsorted(cumprob, np.random.random())
+                if index < cumprob.size:
+                    par[k] = pop[index].copy()
+                else:
+                    par[k] = pop[np.random.randint(0, index - 1)].copy()
                 
-            ef_par_list=np.array([False]*self.par_s)
-            par_count=0
-            while par_count==0:
-                for k in range(0,self.par_s):
-                    if np.random.random()<=self.prob_cross:
-                        ef_par_list[k]=True
-                        par_count+=1
+            ef_par_list = np.full(self.par_s, False) #np.array([False]*self.par_s)
+            par_count = 0
+            while par_count == 0:
+                # it can be vectorized
+                for k in range(0, self.par_s):
+                    if np.random.random() <= self.prob_cross:
+                        ef_par_list[k] = True
+                        par_count += 1
                  
-            ef_par=par[ef_par_list].copy()
+            ef_par = par[ef_par_list].copy()
     
             #############################################################  
             #New generation
-            pop=np.array([np.zeros(self.dim+1)]*self.pop_s)
+            pop = np.empty((self.pop_s, self.dim+1))  #np.array([np.zeros(self.dim+1)]*self.pop_s)
             
-            for k in range(0,self.par_s):
-                pop[k]=par[k].copy()
+            for k in range(0 ,self.par_s):
+                pop[k] = par[k].copy()
                 
             for k in range(self.par_s, self.pop_s, 2):
-                r1=np.random.randint(0,par_count)
-                r2=np.random.randint(0,par_count)
-                pvar1=ef_par[r1,: self.dim].copy()
-                pvar2=ef_par[r2,: self.dim].copy()
+                r1 = np.random.randint(0, par_count)
+                r2 = np.random.randint(0, par_count)
+                pvar1 = ef_par[r1,:self.dim].copy()
+                pvar2 = ef_par[r2,:self.dim].copy()
                 
-                ch=self.cross(pvar1,pvar2,self.c_type)
-                ch1=ch[0].copy()
-                ch2=ch[1].copy()
+                ch1, ch2 = self.cross(pvar1, pvar2, self.c_type)
                 
-                ch1=self.mut(ch1)
-                ch2=self.mutmidle(ch2,pvar1,pvar2)               
-                solo[: self.dim]=ch1.copy()                
-                obj=self.sim(ch1)
-                solo[self.dim]=obj
-                pop[k]=solo.copy()                
-                solo[: self.dim]=ch2.copy()                
-                obj=self.sim(ch2)               
-                solo[self.dim]=obj
-                pop[k+1]=solo.copy()
+                ch1 = self.mut(ch1)
+                ch2 = self.mutmidle(ch2, pvar1, pvar2)               
+                
+                solo[: self.dim] = ch1.copy()                
+                obj = self.sim(ch1)
+                solo[self.dim] = obj
+                pop[k] = solo.copy()                
+                
+                solo[: self.dim] = ch2.copy()                
+                obj = self.sim(ch2)               
+                solo[self.dim] = obj
+                pop[k+1] = solo.copy()
         #############################################################       
-            t+=1
+            t += 1
             if counter > self.mniwi:
                 pop = pop[pop[:,self.dim].argsort()]
-                if pop[0,self.dim]>=self.best_function:
-                    t=self.iterate
-                    self.progress(t,self.iterate,status="GA is running...")
-                    time.sleep(2)
+                if pop[0,self.dim] >= self.best_function:
+                    t = self.iterate # to stop loop
+                    self.progress(t, self.iterate,status="GA is running...")
+                    time.sleep(0.7) #time.sleep(2)
                     t+=1
                     self.stop_mniwi=True
-                
+        
+        
+        
         #############################################################
         #Sort
         pop = pop[pop[:,self.dim].argsort()]
         
-        if pop[0,self.dim]<self.best_function:
+        if pop[0,self.dim] < self.best_function:
                 
-            self.best_function=pop[0,self.dim].copy()
+            self.best_function=pop[0,self.dim]#.copy()
             self.best_variable=pop[0,: self.dim].copy()
         #############################################################
         # Report
@@ -421,17 +422,15 @@ class geneticalgorithm2():
         
         
  
+        self.output_dict={'variable': self.best_variable, 'function': self.best_function}
         
-        self.output_dict={'variable': self.best_variable, 'function':\
-                          self.best_function}
-        
-        show=' '*100
+        show=' '*200
         sys.stdout.write('\r%s' % (show))
         sys.stdout.write('\r The best solution found:\n %s' % (self.best_variable))
         sys.stdout.write('\n\n Objective function:\n %s\n' % (self.best_function))
         sys.stdout.flush() 
         
-        if self.plot:
+        if not no_plot:
             self.plot_results()
 
         if self.stop_mniwi==True:
@@ -455,98 +454,96 @@ class geneticalgorithm2():
         plt.show()
 
 ##############################################################################         
-    def cross(self,x,y,c_type):
+    def cross(self, x, y, c_type):
          
         ofs1=x.copy()
         ofs2=y.copy()
         
 
         if c_type=='one_point':
-            ran=np.random.randint(0,self.dim)
-            for i in range(0,ran):
-                ofs1[i]=y[i].copy()
-                ofs2[i]=x[i].copy()
+            ran=np.random.randint(0, self.dim)
+            ofs1[:ran] = y[:ran]
+            ofs2[:ran] = x[:ran]
   
-        if c_type=='two_point':
+        elif c_type=='two_point':
                 
-            ran1=np.random.randint(0,self.dim)
-            ran2=np.random.randint(ran1,self.dim)
-                
-            for i in range(ran1,ran2):
-                ofs1[i]=y[i].copy()
-                ofs2[i]=x[i].copy()
+            ran1=np.random.randint(0, self.dim)
+            ran2=np.random.randint(ran1, self.dim)
             
-        if c_type=='uniform':
-                
-            for i in range(0, self.dim):
-                ran=np.random.random()
-                if ran <0.5:
-                    ofs1[i]=y[i].copy()
-                    ofs2[i]=x[i].copy() 
+            ofs1[ran1:ran2] = y[ran1:ran2]
+            ofs2[ran1:ran2] = x[ran1:ran2]
+              
+        elif c_type=='uniform':
+            ran = np.random.random(self.dim) < 0.5
+            ofs1[ran] = y[ran]
+            ofs2[ran] = x[ran]
                    
-        return np.array([ofs1,ofs2])
+        
+        return (ofs1, ofs2)
 ###############################################################################  
     
     def mut(self,x):
         
         for i in self.integers[0]:
-            ran=np.random.random()
-            if ran < self.prob_mut:
-                
-                x[i]=np.random.randint(self.var_bound[i][0],\
-                 self.var_bound[i][1]+1) 
+            if np.random.random() < self.prob_mut:
+                x[i]=np.random.randint(self.var_bound[i][0], self.var_bound[i][1]+1) 
                     
         
 
         for i in self.reals[0]:                
-            ran=np.random.random()
-            if ran < self.prob_mut:   
-
-               x[i]=self.var_bound[i][0]+np.random.random()*\
-                (self.var_bound[i][1]-self.var_bound[i][0])    
+            if np.random.random() < self.prob_mut:   
+               x[i]=np.random.uniform(self.var_bound[i][0], self.var_bound[i][1]) #self.var_bound[i][0]+np.random.random()*\
+                #(self.var_bound[i][1]-self.var_bound[i][0])    
             
         return x
+
 ###############################################################################
     def mutmidle(self, x, p1, p2):
         for i in self.integers[0]:
-            ran=np.random.random()
-            if ran < self.prob_mut:
+
+            if np.random.random() < self.prob_mut:
                 if p1[i]<p2[i]:
                     x[i]=np.random.randint(p1[i],p2[i])
                 elif p1[i]>p2[i]:
                     x[i]=np.random.randint(p2[i],p1[i])
                 else:
-                    x[i]=np.random.randint(self.var_bound[i][0],\
-                 self.var_bound[i][1]+1)
+                    x[i]=np.random.randint(self.var_bound[i][0], self.var_bound[i][1]+1)
                         
         for i in self.reals[0]:                
             ran=np.random.random()
             if ran < self.prob_mut:   
-                if p1[i]<p2[i]:
-                    x[i]=p1[i]+np.random.random()*(p2[i]-p1[i])  
-                elif p1[i]>p2[i]:
-                    x[i]=p2[i]+np.random.random()*(p1[i]-p2[i])
+                if p1[i] != p2[i]:
+                    x[i]=np.random.uniform(p1[i], p2[i]) 
                 else:
-                    x[i]=self.var_bound[i][0]+np.random.random()*\
-                (self.var_bound[i][1]-self.var_bound[i][0]) 
+                    x[i]= np.random.uniform(self.var_bound[i][0], self.var_bound[i][1])
         return x
+
+
 ###############################################################################     
     def evaluate(self):
         return self.f(self.temp)
+    
 ###############################################################################    
-    def sim(self,X):
-        self.temp=X.copy()
-        obj=None
+    def sim(self, X):
+        
+        self.temp = X#.copy()
+        
+        obj = None
         try:
-            obj=func_timeout(self.funtimeout,self.evaluate)
+            obj = func_timeout(self.funtimeout, self.evaluate)
         except FunctionTimedOut:
             print("given function is not applicable")
-        assert (obj!=None), "After "+str(self.funtimeout)+" seconds delay "+\
+        
+        assert (obj is not None), "After "+str(self.funtimeout)+" seconds delay "+\
                 "func_timeout: the given function does not provide any output"
+                
+        assert ((type(obj)==int or type(obj)==float or obj.size==1)), "Function should return a number or an np.array with len == 1"
+        
         return obj
 
 ###############################################################################
     def progress(self, count, total, status=''):
+        
         bar_len = 50
         filled_len = int(round(bar_len * count / float(total)))
 
